/*
 * Copyright (c) 1991-1993 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: win32_audio.cc,v 1.1 96/02/04 15:48:53 mccanne Exp $ (LBL)";
#endif

#include <winsock.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "audio.h"
#include "ss.h"
#include "Tcl.h"

class Win32Audio : public Audio {
    public:
	Win32Audio();
	~Win32Audio();
	virtual int FrameReady();
	virtual u_char* Read();
	virtual	void Write(u_char *);
	virtual void SetRGain(int);
	virtual void SetPGain(int);
	virtual void OutputPort(int);
	virtual void InputPort(int);
	virtual void Obtain();
	virtual void Release();
	virtual void RMute();
	virtual void RUnmute();
	virtual int HalfDuplex() const;

    protected:
	int OpenOut();
	void CloseOut();
	int OpenIn();
	void CloseIn();

	HWAVEOUT shWaveOut;
	HWAVEIN  shWaveIn;

	WAVEFORMATEX format;

	HWND hwnd_;

	WAVEHDR whIn;

	int hadIn, hadOut;

	char errorText[MAXERRORLENGTH];

	u_char* writebuf;
	u_char* readbuf;
	u_char* zbuf;

      private:
	static LRESULT CALLBACK AudioHandler(HWND, UINT, WPARAM, LPARAM);
	static void WriteDone(LPWAVEHDR, Win32Audio *);
};

static class Win32AudioMatcher : public Matcher {
public:
	Win32AudioMatcher() : Matcher("audio") {}
	TclObject* match(const char* fmt) {
		if (strcmp(fmt, "pc") == 0)
			return (new Win32Audio);
		return (0);
	}
} win32_audio_matcher;

extern "C" int  dprintf(const char *, ...);
extern "C" HINSTANCE TkWinGetAppInstance();

LRESULT CALLBACK
Win32Audio::AudioHandler(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Win32Audio* p;

	switch (message) {
	case WM_CREATE:{
		CREATESTRUCT *info = (CREATESTRUCT *) lParam;
		SetWindowLong(hwnd, GWL_USERDATA, (DWORD)info->lpCreateParams);
		return 0;
	}	

	case WM_DESTROY:
		return 0;
	
	case MM_WOM_DONE:
		p = (Win32Audio *) GetWindowLong(hwnd, GWL_USERDATA);
		if (p == NULL) {
			dprintf("AudioHandler: no GWL_USERDATA\n");
		}
		WriteDone(LPWAVEHDR(lParam), p);
		return 0;

	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}

Win32Audio::Win32Audio() : shWaveOut(0), shWaveIn(0), hwnd_(0), hadIn(0),
	hadOut(0)
{
	WNDCLASS cl;

	/*
	 * Register the Message window class.
	 */

	cl.style = CS_HREDRAW | CS_VREDRAW;
	cl.lpfnWndProc = AudioHandler;
	cl.cbClsExtra = 0;
	cl.cbWndExtra = 0;
	cl.hInstance = TkWinGetAppInstance();
	cl.hIcon = NULL;
	cl.hCursor = NULL;
	cl.hbrBackground = NULL;
	cl.lpszMenuName = NULL;
	cl.lpszClassName = "Audio";
	RegisterClass(&cl);

	format.wFormatTag = WAVE_FORMAT_MULAW;
	format.nChannels = 1;
	format.nSamplesPerSec = 8000;
	format.wBitsPerSample = 8;
	format.nAvgBytesPerSec =
	  format.nChannels * format.nSamplesPerSec;
	format.nBlockAlign =
	  format.nChannels * format.wBitsPerSample / 8;
	format.cbSize = 0;

	readbuf = new u_char[blksize];
	writebuf = new u_char[blksize];

	zbuf = new u_char[blksize];
	memset(zbuf, ULAW_ZERO, blksize);
}

Win32Audio::~Win32Audio()
{
	CloseIn();
	CloseOut();
	if (hwnd_)
		(void) DestroyWindow(hwnd_);
}

int Win32Audio::HalfDuplex() const
{
	/*XXX change this if full duplex audio device available*/
	return 1;
}

int Win32Audio::OpenOut()
{
	int error = 0;
	DWORD vol;

	if (shWaveOut == 0) {
		error = waveOutOpen(&shWaveOut, WAVE_MAPPER, &format,
				    (DWORD)hwnd_, (DWORD)this,
				    CALLBACK_WINDOW);

		/*
		 * Maybe we failed because someone is playing sound already.
		 * Shut any sound off, and try once more before giving up.
		 */
		if (error) {
			sndPlaySound(NULL, 0);
			error = waveOutOpen(&shWaveOut, WAVE_MAPPER, &format,
					    NULL, 0, CALLBACK_NULL);
			if (error) {
				waveOutGetErrorText(error, errorText, sizeof(errorText));
				fprintf(stderr, "OpenOut: unable to waveOutOpen: %s\n",
					errorText);
				return error;
			}
		}

		error = waveOutGetVolume(shWaveOut, &vol);
		if (error) {
			waveOutGetErrorText(error, errorText, sizeof(errorText));
			fprintf(stderr, "Win32Audio: waveOutGetVolume Error: %s\n",
				errorText);
		}
		else
			pgain = vol & 0xff;
	}
	return error;
}

int Win32Audio::OpenIn()
{
	int error = 0;

	if (shWaveIn == 0) {
		error = waveInOpen(&shWaveIn, WAVE_MAPPER, &format,
				   0, 0, CALLBACK_NULL);
		if (error == MMSYSERR_NOERROR) {
			whIn.lpData = (LPSTR) readbuf;
			whIn.dwBufferLength = blksize;
			whIn.dwFlags = 0;
			error = waveInPrepareHeader(shWaveIn, &whIn, sizeof(whIn));
			if (error) {
				waveInGetErrorText(error, errorText, sizeof(errorText));
				fprintf(stderr, "Win32Audio: waveInPrepareHeader: (%d) %s\n",
					error, errorText);
			}
			error = waveInAddBuffer(shWaveIn, &whIn, sizeof(whIn));
			if (error) {
				waveInGetErrorText(error, errorText, sizeof(errorText));
				fprintf(stderr, "Win32Audio: waveInAddBuffer: (%d) %s\n",
					error, errorText);
			}

		}
		else {
			waveInGetErrorText(error, errorText, sizeof(errorText));
			fprintf(stderr, "Win32Audio: waveInOpen Error: (%d) %s\n",
				error, errorText);
		}
	}
	return error;
}

void Win32Audio::CloseOut()
{
	if (shWaveOut) {
		(void) waveOutReset(shWaveOut);
		(void) waveOutClose(shWaveOut);
		shWaveOut = 0;
	}
}

void Win32Audio::CloseIn()
{
	if (shWaveIn) {
		(void) waveInReset(shWaveIn);
		(void) waveInUnprepareHeader(shWaveIn, &whIn, sizeof(whIn));
		(void) waveInClose(shWaveIn);
		shWaveIn = 0;
	}
}

void Win32Audio::Obtain()
{
	int error;

	if (HaveAudio())
		abort();

	if (hwnd_ == 0) {
		hwnd_ = CreateWindow("Audio", "",
				     WS_POPUP | WS_CLIPCHILDREN,
				     CW_USEDEFAULT, CW_USEDEFAULT, 1, 1,
				     NULL,
				     NULL, TkWinGetAppInstance(), this);
		ShowWindow(hwnd_, SW_HIDE);
	}

	if (hadIn && (shWaveOut == 0))
		error = OpenIn();
	else if (hadOut && (shWaveIn == 0))
		error = OpenOut();
	else if (shWaveOut == 0)
		error = OpenOut();

	hadIn = hadOut = 0;

	fd = (error)?-1:0;

#ifdef fullduplex
	Audio::Obtain();
#else
	notify();
#endif
}

void Win32Audio::Release()
{
	dprintf("Release: out=%x in=%x\n", shWaveOut, shWaveIn);

	hadIn = (shWaveIn != 0);
	hadOut = (shWaveOut != 0);
	CloseOut();
	CloseIn();
	Audio::Release();
}

void Win32Audio::WriteDone(LPWAVEHDR lpWh, Win32Audio *lpAudio)
{
	if (lpWh->dwFlags & WHDR_DONE) {
		(void) waveOutUnprepareHeader(lpAudio->shWaveOut, lpWh,
					      sizeof(WAVEHDR));
		delete [] LPWAVEHDR(lpWh);
	}
	else
		dprintf("WriteDone: buffer not done. dwFlags=%x\n",
			lpWh->dwFlags);
}

void Win32Audio::Write(u_char *cp)
{
	int error;
	LPWAVEHDR whp;

	if ((rmute & 1) != 0) {

		whp = new WAVEHDR;
		whp->lpData = (LPSTR) cp;
		whp->dwBufferLength = blksize;
		whp->dwFlags = 0;
		error = waveOutPrepareHeader(shWaveOut, whp, sizeof(WAVEHDR));
		if (error) {
			waveOutGetErrorText(error, errorText, sizeof(errorText));
			fprintf(stderr, "Win32Audio: waveOutPrepareHeader: %s\n",
				errorText);
			return;
		}

		error = waveOutWrite(shWaveOut, whp, sizeof(WAVEHDR));
		if (error) {
			waveOutGetErrorText(error, errorText, sizeof(errorText));
			fprintf(stderr, "Win32Audio: waveOutWrite: %s\n",
				errorText);
			return;
		}
	}
}

int Win32Audio::FrameReady()
{
	if ((rmute & 1) != 0)
		return (1);

	return ((whIn.dwFlags & WHDR_DONE)?1:0);
}

u_char* Win32Audio::Read()
{
	int error;
	u_char* cp;

	if ((rmute & 1) != 0)
		cp = zbuf;
	else {
		if (whIn.dwBytesRecorded != (DWORD)blksize)
			dprintf("Read: audio read count=%d (blksize %d)\n",
				whIn.dwBytesRecorded, blksize);

		cp = readbuf;
		error = waveInAddBuffer(shWaveIn, &whIn, sizeof(whIn));
		if (error) {
			waveInGetErrorText(error, errorText, sizeof(errorText));
			fprintf(stderr, "Win32Audio: waveInAddBuffer: (%d) %s\n",
				error, errorText);
		}
	}
	return (cp);
}

void Win32Audio::SetRGain(int level)
{
	rgain = level;
}

void Win32Audio::SetPGain(int level)
{
	DWORD vol;
	int error;

	pgain = level;
	if (level >= 255)
		level = (short)-1;
	else
		level <<= 8;
	vol = level | (level << 16);

	error = waveOutSetVolume(shWaveOut, vol);
	if (error) {
		waveOutGetErrorText(error, errorText, sizeof(errorText));
		fprintf(stderr, "Win32Audio: waveOutSetVolume: %s\n",
			errorText);
	}
}

void Win32Audio::OutputPort(int p)
{
	oport = p;
}

void Win32Audio::InputPort(int p)
{
	iport = p;
}

void Win32Audio::RMute()
{
	(void) waveInStop(shWaveIn);

	CloseIn();
	if (OpenOut())
		return;

	rmute |= 1;
}

void Win32Audio::RUnmute()
{
	CloseOut();
	if (OpenIn())
		return;

	int error = waveInStart(shWaveIn);
	if (error) {
		waveInGetErrorText(error, errorText, sizeof(errorText));
		fprintf(stderr, "Win32Audio: waveInStart: (%d) %s\n",
			error, errorText);
		return;
	}
	rmute &=~ 1;
}
